import asyncio
from functools import wraps
from typing import Any, Callable, Dict, Optional, Union

from ..backends.interface import Backend
from ..key import get_cache_key, get_cache_key_template, get_call_values, get_func_params
from ..typing import FuncArgsType
from .defaults import CacheDetect, _default_store_condition, context_cache_detect

__all__ = ("cache", "invalidate")


def cache(
    backend: Backend,
    ttl: int,
    func_args: FuncArgsType = None,
    key: Optional[str] = None,
    store: Optional[Callable[[Any], bool]] = None,
    prefix: str = "",
):
    """
    Simple cache strategy - trying to return cached result,
    execute wrapped call and store a result with ttl if condition return true
    :param backend: cache backend
    :param ttl: duration in seconds to store a result
    :param func_args: arguments that will be used in key
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param store: callable object that determines whether the result will be saved or not
    :param prefix: custom prefix for key
    """
    store = _default_store_condition if store is None else store

    def _decor(func):
        func._key_template = prefix + get_cache_key_template(func, func_args=func_args, key=key)

        @wraps(func)
        async def _wrap(*args, _from_cache: CacheDetect = context_cache_detect, **kwargs):
            _cache_key = prefix + get_cache_key(func, args, kwargs, func_args, key)
            cached = await backend.get(_cache_key)
            if cached:
                _from_cache.set(_cache_key, ttl=ttl)
                return cached
            result = await func(*args, **kwargs)
            if store(result):
                await backend.set(_cache_key, result, expire=ttl)
            return result

        return _wrap

    return _decor


async def invalidate_func(backend: Backend, func, kwargs: Optional[Dict] = None):
    key_template = getattr(func, "_key_template", None)
    if not key_template:
        return None
    values = {**{param: "*" for param in get_func_params(func)}, **kwargs}
    values = {k: str(v) if v is not None else "" for k, v in values.items()}
    return await backend.delete_match(key_template.format(**values).lower())


def invalidate(
    backend: Backend,
    target: Union[str, Callable],
    args_map: Optional[Dict[str, str]] = None,
    defaults: Optional[Dict[str, Any]] = None,
):
    args_map = args_map or {}
    defaults = defaults or {}

    def _decor(func):
        @wraps(func)
        async def _wrap(*args, **kwargs):
            result = await func(*args, **kwargs)
            _args = get_call_values(func, args, kwargs, func_args=None)
            _args.update(defaults)
            for source, dest in args_map.items():
                if dest in _args:
                    _args[source] = _args.pop(dest)
                if callable(dest):
                    _args[source] = dest(*args, **kwargs)
            if callable(target):
                asyncio.create_task(invalidate_func(backend, target, _args))
            else:
                asyncio.create_task(
                    backend.delete_match(target.format({k: str(v) if v is not None else "" for k, v in _args.items()}))
                )
            return result

        return _wrap

    return _decor



from datetime import datetime, timedelta
from functools import wraps
from typing import Optional, Tuple, Type, Union

from ..backends.interface import Backend
from ..key import get_cache_key
from ..typing import TTL, FuncArgsType

__all__ = ("circuit_breaker", "CircuitBreakerOpen")


class CircuitBreakerOpen(Exception):
    pass


def circuit_breaker(
    backend: Backend,
    errors_rate: int,
    period: TTL,
    ttl: int,
    exceptions: Union[Type[Exception], Tuple[Type[Exception]]] = Exception,
    key: Optional[str] = None,
    func_args: FuncArgsType = None,
    prefix: str = "circuit_breaker",
):
    """
    Circuit breaker
    :param backend: cache backend
    :param errors_rate: Errors rate in percents
    :param period: Period
    :param ttl: seconds in int or as timedelta to keep circuit breaker switched
    :param func_args: arguments that will be used in key
    :param exceptions: exceptions at which returned cache result
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param prefix: custom prefix for key, default "circuit_breaker"
    """
    assert 0 < errors_rate < 100

    def _decor(func):
        @wraps(func)
        async def _wrap(*args, **kwargs):
            _period = period() if callable(period) else period
            _period = _period.total_seconds() if isinstance(_period, timedelta) else _period
            _cache_key = prefix + ":" + get_cache_key(func, args, kwargs, func_args, key)
            if await backend.is_locked(_cache_key + ":open"):
                raise CircuitBreakerOpen()
            bucket = _get_bucket_number(_period, segments=100)
            total_in_bucket = await backend.incr(_cache_key + f":total:{bucket}")
            if total_in_bucket == 1:
                await backend.expire(key=_cache_key + f":total:{bucket}", timeout=_period)
                await backend.set(key=_cache_key + ":fails", value=0, expire=_period)
            try:
                return await func(*args, **kwargs)
            except exceptions as exc:
                fails = await backend.incr(_cache_key + ":fails")
                total = total_in_bucket + await _get_buckets_values(backend, segments=100, except_number=bucket)
                if fails * 100 / total >= errors_rate:
                    await backend.set_lock(_cache_key + ":open", True, expire=ttl)
                raise exc

        return _wrap

    return _decor


def _get_bucket_number(period: Union[int, timedelta], segments: int) -> int:
    return int((datetime.utcnow().timestamp() % period) / segments)


async def _get_buckets_values(backend: Backend, segments: int, except_number: int) -> int:
    keys = list(range(segments))
    keys.remove(except_number)
    return sum([v for v in await backend.get_many(*keys) if v])



import asyncio
import datetime
import re
from typing import Any, Optional, Tuple, Union

from .interface import Backend

__all__ = ("Memory", "MemoryInterval")


class Memory(Backend):
    def __init__(self, size: int = 1000):
        self.store = {}
        self.size = size
        self._meta = {}
        self._lock = asyncio.Lock()
        super().__init__()

    async def clear(self):
        self.store = {}

    async def set(self, key: str, value: Any, expire: Union[None, float, int] = None, exist=None) -> bool:
        if exist is not None:
            if not (key in self.store) is exist:
                return False
        async with self._lock:
            self._set(key, value, expire)
        return True

    async def get(self, key: str) -> Any:
        return self._get(key)

    async def get_many(self, *keys: str) -> Tuple:
        return tuple([self._get(key) for key in keys])

    async def incr(self, key: str):
        value = int(self._get(key) or 0) + 1
        self._set(key=key, value=value)
        return value

    async def delete(self, key: str):
        async with self._lock:
            self._delete(key)

    def _delete(self, key: str) -> bool:
        if key in self.store:
            del self.store[key]
            if key in self._meta:
                del self._meta[key]
            return True
        return False

    async def delete_match(self, pattern: str):
        pattern = pattern.replace("*", "[^:]+")
        regexp = re.compile(pattern)
        for key in dict(self.store):
            if regexp.fullmatch(key):
                await self.delete(key)

    async def expire(self, key: str, timeout: float):
        value = self._get(key)
        if value is None:
            return
        self._set(key=key, value=value, expire=timeout)

    async def get_expire(self, key: str) -> int:
        return -1

    async def ping(self, message: Optional[str] = None):
        return b"PONG" if message is None else message

    def _set(self, key: str, value: Any, expire: Optional[float] = None):
        if len(self.store) > self.size:
            return
        self.store[key] = value

        if expire is not None and expire > 0.0:
            loop = asyncio.get_event_loop()
            handler = self._meta.get(key)
            if handler:
                handler.cancel()
            self._meta[key] = loop.call_later(expire, self._delete, key)

    def _get(self, key: str) -> Optional[Any]:
        return self.store.get(key, None)

    async def set_lock(self, key: str, value, expire):
        return await self.set(key, value, expire=expire, exist=False)

    async def is_locked(self, key: str, wait=None, step=0.1) -> bool:
        if wait is None:
            return key in self.store
        while wait > 0:
            if key not in self.store:
                return False
            wait -= step
            await asyncio.sleep(step)
        return key in self.store

    async def unlock(self, key, value) -> bool:
        return self._delete(key)


class MemoryInterval(Memory):
    def __init__(self, size: int = 1000, check_interval: float = 1.0):
        self._check_interval = check_interval
        super().__init__(size=size)

    async def init(self):
        asyncio.create_task(self._remove_expired())

    async def _remove_expired(self):
        while True:
            store = dict(self.store)
            for key in store:
                await self.get(key)
            await asyncio.sleep(self._check_interval)

    async def get(self, key: str) -> Optional[Any]:
        expiration = self._meta.get(key)
        if expiration and datetime.datetime.utcnow() > expiration:
            await self.delete(key)
            return None

        return self.store.get(key, None)

    def _set(self, key: str, value: Any, expire: Optional[float] = None) -> None:
        self.store[key] = value

        if expire is not None and expire > 0.0:
            self._meta[key] = datetime.datetime.utcnow() + datetime.timedelta(seconds=expire)

    async def get_expire(self, key: str) -> int:
        if key in self._meta:
            return abs(int((datetime.datetime.utcnow() - self._meta[key]).total_seconds()))
        return -1
